@@ -1,9 +1,9 @@
-plot.rc <- function(x, dim=c(1, 2), what=c("both", "rows", "columns"),
+plot.rc <- function(x, dim=c(1, 2), what=c("both", "rows", "columns"), which=NULL,
                     mass=TRUE, luminosity=length(x$assoc$diagonal > 0),
                     conf.ellipses=NA, replicates=FALSE,
                     coords=c("cartesian", "polar"), rev.axes=c(FALSE, FALSE),
                     cex=par("cex"), col=c("blue", "red"), groups=NULL,
-                    xlim, ylim, asp, xlab, ylab, main, ...) {
+                    xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   what <- match.arg(what)
   coords <- match.arg(coords)
 
@@ -13,18 +13,19 @@ plot.rc <- function(x, dim=c(1, 2), what=c("both", "rows", "columns"),
   if(!length(x$assoc) > 0)
       stop("x must contain an association component")
 
-  plot.assoc(x$assoc, dim=dim, what=what, mass=mass, luminosity=luminosity,
+  plot.assoc(x$assoc, dim=dim, what=what, which=which, mass=mass, luminosity=luminosity,
              arrow=NULL, conf.ellipses=conf.ellipses, replicates=replicates,
              coords=coords, rev.axes=rev.axes, cex=cex, col=col, groups=groups,
-             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, ...)
+             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main,
+             pch=pch, font=font, ...)
 }
 
-plot.rc.symm <- function(x, dim=c(1, 2),
+plot.rc.symm <- function(x, dim=c(1, 2), which=NULL,
                          mass=TRUE, luminosity=length(x$assoc$diagonal > 0),
                          conf.ellipses=NA, replicates=FALSE,
                          coords=c("cartesian", "polar"), rev.axes=c(FALSE, FALSE),
                          cex=par("cex"), col="blue", groups=NULL,
-                         xlim, ylim, asp, xlab, ylab, main, ...) {
+                         xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   coords <- match.arg(coords)
 
   if(!inherits(x, "rc.symm"))
@@ -33,18 +34,19 @@ plot.rc.symm <- function(x, dim=c(1, 2),
   if(!length(x$assoc) > 0)
       stop("x must contain an association component")
 
-  plot.assoc(x$assoc, dim=dim, what="rows", mass=mass, luminosity=luminosity,
+  plot.assoc(x$assoc, dim=dim, what="rows", which=which, mass=mass, luminosity=luminosity,
              arrow=NULL, conf.ellipses=conf.ellipses, replicates=replicates,
              coords=coords, rev.axes=rev.axes, cex=cex, col=col, groups=groups,
-             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, ...)
+             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main,
+             pch=pch, font=font, ...)
 }
 
-plot.hmskew <- function(x, dim=c(1, 2), what=c("skew-symmetric", "symmetric"),
+plot.hmskew <- function(x, dim=c(1, 2), what=c("skew-symmetric", "symmetric"), which=NULL,
                         mass=TRUE, luminosity=length(x$assoc.hmskew$diagonal > 0), arrow=45,
                         conf.ellipses=NA, replicates=FALSE,
                         coords=c("polar", "cartesian"), rev.axes = c(FALSE, FALSE),
                         cex=par("cex"), col="blue", groups=NULL,
-                        xlim, ylim, asp, xlab, ylab, main, ...) {
+                        xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   if(!inherits(x, "hmskew"))
       stop("x must be a hmskew object")
 
@@ -66,18 +68,19 @@ plot.hmskew <- function(x, dim=c(1, 2), what=c("skew-symmetric", "symmetric"),
   if(what == "skew-symmetric" && missing(ylab))
       ylab <- ""
 
-  plot.assoc(ass, dim=dim, what="rows", mass=mass, luminosity=luminosity,
+  plot.assoc(ass, dim=dim, what="rows", which=which, mass=mass, luminosity=luminosity,
              arrow=arrow, conf.ellipses=conf.ellipses, replicates=replicates,
              coords=coords, rev.axes=rev.axes, cex=cex, col=col, groups=groups,
-             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, ...)
+             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main,
+             pch=pch, font=font, ...)
 }
 
-plot.yrcskew <- function(x, dim=c(1, 2), what=c("skew-symmetric", "symmetric"),
+plot.yrcskew <- function(x, dim=c(1, 2), what=c("skew-symmetric", "symmetric"), which=NULL,
                          mass=TRUE, luminosity=length(x$assoc.yrcskew$diagonal > 0), arrow=45,
                          conf.ellipses=NA, replicates=FALSE,
                          coords=c("polar", "cartesian"), rev.axes = c(FALSE, FALSE),
                          cex=par("cex"), col="blue", groups=NULL,
-                         xlim, ylim, asp, xlab, ylab, main, ...) {
+                         xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   if(!inherits(x, "yrcskew"))
       stop("x must be a yrcskew object")
 
@@ -98,18 +101,19 @@ plot.yrcskew <- function(x, dim=c(1, 2), what=c("skew-symmetric", "symmetric"),
   if(what == "skew-symmetric" && missing(ylab))
       ylab <- ""
 
-  plot.assoc(ass, dim=dim, what="rows", mass=mass, luminosity=luminosity,
+  plot.assoc(ass, dim=dim, what="rows", which=which, mass=mass, luminosity=luminosity,
              arrow=arrow, conf.ellipses=conf.ellipses, replicates=replicates,
              coords=coords, rev.axes=rev.axes, cex=cex, col=col, groups=groups,
-             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, ...)
+             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main,
+             pch=pch, font=font, ...)
 }
 
-plot.rcL <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"),
+plot.rcL <- function(x, dim=c(1, 2), layer="average", what=c("both", "rows", "columns"), which=NULL,
                     mass=TRUE, luminosity=length(x$assoc$diagonal > 0),
                     conf.ellipses=NA, replicates=FALSE,
                     coords=c("cartesian", "polar"), rev.axes=c(FALSE, FALSE),
                     cex=par("cex"), col=c("blue", "red"), groups=NULL,
-                    xlim, ylim, asp, xlab, ylab, main, ...) {
+                    xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   what <- match.arg(what)
   coords <- match.arg(coords)
 
@@ -119,18 +123,19 @@ plot.rcL <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"),
   if(!length(x$assoc) > 0)
       stop("x must contain an association component")
 
-  plot.assoc(x$assoc, dim=dim, layer=layer, what=what, mass=mass, luminosity=luminosity,
+  plot.assoc(x$assoc, dim=dim, layer=layer, what=what, which=which, mass=mass, luminosity=luminosity,
              arrow=NULL, conf.ellipses=conf.ellipses, replicates=replicates,
              coords=coords, rev.axes=rev.axes, cex=cex, col=col, groups=groups,
-             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, ...)
+             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main,
+             pch=pch, font=font, ...)
 }
 
-plot.rcL.symm <- function(x, dim=c(1, 2), layer=1,
+plot.rcL.symm <- function(x, dim=c(1, 2), layer="average", which=NULL,
                           mass=TRUE, luminosity=length(x$assoc$diagonal > 0),
                           conf.ellipses=NA, replicates=FALSE,
                           coords=c("cartesian", "polar"), rev.axes=c(FALSE, FALSE),
                           cex=par("cex"), col="blue", groups=NULL,
-                          xlim, ylim, asp, xlab, ylab, main, ...) {
+                          xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   coords <- match.arg(coords)
 
   if(!inherits(x, "rcL.symm"))
@@ -139,19 +144,20 @@ plot.rcL.symm <- function(x, dim=c(1, 2), layer=1,
   if(!length(x$assoc) > 0)
       stop("x must contain an association component")
 
-  plot.assoc(x$assoc, dim=dim, layer=layer, what="rows", mass=mass, luminosity=luminosity,
+  plot.assoc(x$assoc, dim=dim, layer=layer, what="rows", which=which, mass=mass, luminosity=luminosity,
              arrow=NULL, conf.ellipses=conf.ellipses, replicates=replicates,
              coords=coords, rev.axes=rev.axes, cex=cex, col=col, groups=groups,
-             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, ...)
+             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main,
+             pch=pch, font=font, ...)
 }
 
 
-plot.hmskewL <- function(x, dim=c(1, 2), layer=1, what=c("skew-symmetric", "symmetric"),
+plot.hmskewL <- function(x, dim=c(1, 2), layer="average", what=c("skew-symmetric", "symmetric"), which=NULL,
                          mass=TRUE, luminosity=length(x$assoc.hmskew$diagonal > 0), arrow=45,
                          conf.ellipses=NA, replicates=FALSE,
                          coords=c("polar", "cartesian"), rev.axes=c(FALSE, FALSE),
                          cex=par("cex"), col="blue", groups=NULL,
-                         xlim, ylim, asp, xlab, ylab, main, ...) {
+                         xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   if(!inherits(x, "hmskewL"))
       stop("x must be a hmskewL object")
 
@@ -173,18 +179,19 @@ plot.hmskewL <- function(x, dim=c(1, 2), layer=1, what=c("skew-symmetric", "symm
   if(what == "skew-symmetric" && missing(ylab))
       ylab <- ""
 
-  plot.assoc(ass, dim=dim, layer=layer, what="rows", mass=mass, luminosity=luminosity,
+  plot.assoc(ass, dim=dim, layer=layer, what="rows", which=which, mass=mass, luminosity=luminosity,
              arrow=arrow, conf.ellipses=conf.ellipses, replicates=replicates,
              coords=coords, rev.axes=rev.axes, cex=cex, col=col, groups=groups,
-             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, ...)
+             xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab,
+             main=main, font=font, ...)
 }
 
 plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"),
-                       mass=TRUE, luminosity=length(x$diagonal > 0), arrow=NULL,
+                       which=NULL, mass=TRUE, luminosity=length(x$diagonal > 0), arrow=NULL,
                        conf.ellipses=NA, replicates=FALSE,
                        coords=c("cartesian", "polar"), rev.axes=c(FALSE, FALSE),
                        cex=par("cex"), col=c("blue", "red"), groups=NULL,
-                       xlim, ylim, asp, xlab, ylab, main, ...) {
+                       xlim, ylim, asp, xlab, ylab, main, pch, font, ...) {
   if(!(inherits(x, "assoc")))
       stop("x must be an assoc object")
 
@@ -198,15 +205,15 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
   if(any(dim > ncol(x$row)))
       stop("dim must be a valid dimension of the model")
 
-  if(is.matrix(x$phi) &&
+  if(is.matrix(x$phi) && !layer %in% c("average", "average.rotate") &&
      ((is.numeric(layer) && layer > nrow(x$phi)) ||
       (!is.numeric(layer) && !layer %in% rownames(x$phi))))
       stop("layer must be a valid layer of the model")
 
-  if(!is.numeric(layer))
+  if(!layer %in% c("average", "average.rotate") && !is.numeric(layer))
       layer <- match(layer, rownames(x$phi))
 
-  if(is.matrix(x$phi))
+  if(!layer %in% c("average", "average.rotate") && is.matrix(x$phi))
       layer.name <- rownames(x$phi[layer,, drop=FALSE])
   else
       layer.name <- ""
@@ -218,6 +225,10 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
   nr <- nrow(x$row)
   nc <- nrow(x$col)
 
+  probs <- get.probs(x)
+  rp <- probs$rp
+  cp <- probs$cp
+
   rev.axes <- rep(rev.axes, length.out=2)
 
   if(!is.na(conf.ellipses) && !isTRUE(conf.ellipses > 0 && conf.ellipses < 1))
@@ -240,69 +251,144 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
   what <- match.arg(what)
   coords <- match.arg(coords)
 
-  # Plotting only uses one layer, so get rid of others to make code cleaner below
-  x$phi <- x$phi[layer,]
+  if(inherits(x, "assoc.symm")) {
+       #stopifnot(identical(x$row, x$col))
+       what <- "rows"
+  }
 
-  if(dim(x$row)[3] > 1)
-      x$row <- x$row[,,layer]
-  else
-      x$row <- x$row[,,1]
+  rot <- NULL
+  if(layer %in% c("average", "average.rotate")) {
+      # For homogeneous association with layer, compute a weighted average of phi over layers
+      # And if layer="average.rotate", prepare the drawing of lines representing the axes with the highest variance
 
-  if(dim(x$col)[3] > 1)
-      x$col <- x$col[,,layer]
-  else
-      x$col <- x$col[,,1]
+      if(nl == 1 || nlr > 1 || nlc > 1)
+          stop("'layer=\"average\"' and 'layer=\"average.rotate\"' is only supported with homogeneous layer effect")
+
+      res <- averaged.assoc(x, type=layer)
+
+      rot <- res$rot
+      x$phi <- res$phi
+      x$row <- res$row
+      x$col <- res$col
+  }
+  else {
+      # Plotting only uses one layer, so get rid of others to make code cleaner below
+      x$phi <- x$phi[layer,]
+
+      if(nlr > 1)
+          x$row <- x$row[,,layer]
+      else
+          x$row <- x$row[,,1]
+
+      if(nlc > 1)
+          x$col <- x$col[,,layer]
+      else
+          x$col <- x$col[,,1]
+  }
 
   if(isTRUE(nrow(x$diagonal) > 1))
       x$diagonal <- x$diagonal[layer,]
   else if(length(x$diagonal) > 0)
       x$diagonal <- x$diagonal[1,]
 
-  if(inherits(x, "assoc.symm")) {
-       stopifnot(identical(x$row, x$col))
-       sc <- x$row
-       p <- x$row.weights
+  if(what != "both") {
+      if(is.null(which))
+          which <- TRUE
+
+      if(is.logical(which))
+          which <- seq.int(1, nr)[which]
+      else if(is.character(which) && what == "rows")
+          which <- match(which, rownames(x$row), nomatch=0)
+      else if(is.character(which) && what == "columns")
+          which <- match(which, rownames(x$col), nomatch=0)
+      else if(!isTRUE(all(which >= 1 & which <= nr)))
+          stop("Value of 'which' is invalid.")
+
+      if(what == "rows")
+          which <- list(which, FALSE)
+      else
+          which <- list(FALSE, which)
   }
-  else if(what == "rows") {
-       sc <- x$row
-       p <- x$row.weights
+  else {
+      if(is.null(which))
+          which <- list(TRUE, TRUE)
+      else if(!is.list(which) || length(which) != 2)
+          stop("'which' must be a list with exactly two components when 'what == \"both\"'.")
+
+      if(is.logical(which[[1]]))
+          which[[1]] <- seq.int(1, nr)[which[[1]]]
+      else if(is.character(which[[1]]))
+          which[[1]] <- match(which[[1]], rownames(x$row))
+      else if(!isTRUE(all(which[[1]] >= 1 & which[[1]] <= nr)))
+          stop("First component of 'which' is invalid.")
+
+      if(is.logical(which[[2]]))
+          which[[2]] <- seq.int(1, nc)[which[[2]]]
+      else if(is.character(which[[2]]))
+          which[[2]] <- match(which[[2]], rownames(x$col))
+      else if(!isTRUE(all(which[[2]] >= 1 & which[[2]] <= nc)))
+          stop("Second component of 'which' is invalid.")
+  }
+
+  nwr <- length(which[[1]])
+  nwc <- length(which[[2]])
+
+  if(what == "rows") {
+       sc <- x$row[which[[1]],, drop=FALSE]
+       p <- rp[which[[1]]]
        if(length(col) == 2)
            col <- col[1]
   }
   else if(what == "columns") {
-       sc <- x$col
-       p <- x$col.weights
+       sc <- x$col[which[[2]],, drop=FALSE]
+       p <- cp[which[[2]]]
        if(length(col) == 2)
            col <- col[2]
   }
   else {
-       sc <- rbind(x$row, x$col)
-       p <- c(x$row.weights, x$col.weights)
+       sc <- rbind(x$row[which[[1]],, drop=FALSE], x$col[which[[2]],, drop=FALSE])
+       p <- c(rp[which[[1]]], cp[which[[2]]])
 
        if(length(col) == 2)
-           col <- c(rep(col[1], nr), rep(col[2], nr))
+           col <- c(rep(col[1], nwr), rep(col[2], nwc))
 
        if(length(groups) == 0)
-           groups <- c(rep(2, nr), rep(1, nr))
+           groups <- c(rep(2, nwr), rep(1, nwc))
   }
 
   nsc <- nrow(sc)
 
+  if(nsc == 0)
+      stop("Values of 'what' and 'which' combined do not retain any point to plot.")
+
   if(length(col) == 0)
       col <- "black"
 
-  if(length(groups) != nsc)
+  if(length(groups) > 0 && length(groups) != nsc)
       groups <- rep(groups, length=nsc)
 
+  if(length(groups) > 0 && !is.numeric(groups))
+      groups <- factor(groups)
+
   if(length(col) != nsc)
       col <- rep(col, length=nsc)
 
   if(length(cex) != nsc)
       cex <- rep(cex, length=nsc)
 
+  if(missing(pch) && length(groups) > 0)
+      pch <- rep(c(21, 24, 22, 23, 25), length.out=8)[groups]
+  else if(missing(pch))
+      pch <- rep(21, length.out=nsc)
+
+  if(missing(font))
+      font <- rep(1, length.out=nsc)
+  else if(length(font) != nsc)
+      font <- rep(font, length.out=nsc)
+
   # Integrate phi to scores for graphical representation
   # Cf. Wong (2010), eq. 2.17 and 2.38, or Clogg & Shihadeh (1994), p. 91
-  sc[,dim] <- sweep(sc[,dim], 2, sqrt(abs(x$phi[dim])), "*")
+  sc[, dim] <- sweep(sc[, dim, drop=FALSE], 2, sqrt(abs(x$phi[dim])), "*")
 
   # If phi is negative, change sign of columns so that the interpretation
   # is consistent with positive phi
@@ -310,19 +396,19 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
   # find.stable.scores() and find.stable.scores.hmskew() do the same and wee need to be consistent
   if(!inherits(x, "assoc.symm")) {
       if(what == "columns")
-          sc[,dim] <- sweep(sc[,dim], 2, sign(x$phi[dim]), "*")
+          sc[, dim] <- sweep(sc[, dim, drop=FALSE], 2, sign(x$phi[dim]), "*")
       else if(what == "both")
-          sc[-(1:nr),dim] <- sweep(sc[-(1:nr),dim], 2, sign(x$phi[dim]), "*")
+          sc[-(1:nwr), dim] <- sweep(sc[-(1:nwr), dim, drop=FALSE], 2, sign(x$phi[dim]), "*")
 
       # For printing below
       x$phi[dim] <- abs(x$phi[dim])
   }
 
   if(isTRUE(rev.axes[1]))
-      sc[,dim[1]] <- -sc[,dim[1]]
+      sc[, dim[1]] <- -sc[, dim[1]]
 
   if(isTRUE(rev.axes[2]))
-      sc[,dim[2]] <- -sc[,dim[2]]
+      sc[, dim[2]] <- -sc[, dim[2]]
 
   if(missing(xlim))
       xlim <- range(sc[,dim[1]])
@@ -338,21 +424,44 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
 
   if(coords == "cartesian") {
       if(missing(xlab))
-          xlab <- sprintf("Dimension %i (%.2f)", dim[1], x$phi[dim[1]])
+          xlab <- sprintf("Dimension %i (%s)",
+                          dim[1], prettyNum(round(x$phi[dim[1]], 2), nsmall=2))
 
       if(missing(ylab))
-          ylab <- sprintf("Dimension %i (%.2f)", dim[2], x$phi[dim[2]])
+          ylab <- sprintf("Dimension %i (%s)",
+                          dim[2], prettyNum(round(x$phi[dim[2]], 2), nsmall=2))
 
-      plot(sc[,dim], xlim=xlim, ylim=ylim, asp=asp, xlab=xlab, ylab=ylab, main=main, type="n", ...)
+      plot(sc[, dim, drop=FALSE], xlim=xlim, ylim=ylim, asp=asp,
+           xlab=xlab, ylab=ylab, main=main, type="n", ...)
 
       abline(h=0, lty="dotted")
       abline(v=0, lty="dotted")
+
+      if(!is.null(rot)) {
+          uvrot <- diag(1, nd, nd) %*% rot
+
+          for(i in 1:nd) {
+              coord <- uvrot[i,dim[2]]/uvrot[i, dim[1]]
+              pos <- if(abs(coord) > .5) 2 else 1
+              abline(0, coord, lty="dotted", col="dark grey", lwd=2)
+
+              if(coord > 0)
+                  lim <- min(abs(par("usr")[2]), abs(par("usr")[4]))
+              else
+                  lim <- min(abs(par("usr")[2]), abs(par("usr")[3]))
+
+              if(abs(coord) > 1)
+                  text(-.8 * 1/coord * lim, -.8 * lim, label=paste("Dim", i), col="dark grey", pos=pos)
+              else
+                  text(.8 * lim, .8 * coord * lim, label=paste("Dim", i), col="dark grey", pos=pos)
+          }
+      }
   }
   else {
       opar <- par(mar=c(1, 1, 1, 1))
       on.exit(par(opar))
 
-      plot(sc[,dim], xlim=xlim, ylim=ylim, asp=asp, xlab="", ylab="", main=main,
+      plot(sc[, dim, drop=FALSE], xlim=xlim, ylim=ylim, asp=asp, xlab="", ylab="", main=main,
            xaxt="n", yaxt="n", type="n", bty="n", ...)
 
       abline(h=0, col="grey")
@@ -368,31 +477,30 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
   }
 
   if(!is.na(conf.ellipses)) {
+      if(layer == "average.rotate")
+          stop("Plotting confidence ellipses is not supported when 'layer=\"average.rotate\"'")
+
       covmat <- x$adj.covmats[,, layer]
 
       i <- 0
       start <- c((dim[1] - 1) * (nr + nc), (dim[2] - 1) * (nr + nc))
 
       if(what %in% c("rows", "both")) {
-
-
-          for(i in 1:nr)
-              polygon(ellipse(covmat[start + i, start + i], centre=sc[i, dim], level=conf.ellipses),
+          for(i in 1:nwr)
+              polygon(ellipse(covmat[start + which[[1]][i], start + which[[1]][i]],
+                              centre=sc[i, dim], level=conf.ellipses),
                       border="dark grey", lty="dashed", lwd=2)
       }
 
       if(what %in% c("columns", "both")) {
-          for(j in 1:nc)
-              polygon(ellipse(covmat[start + nr + j, start + nr + j], centre=sc[i + j, dim], level=conf.ellipses),
+          for(j in 1:nwc)
+              polygon(ellipse(covmat[start + nr + which[[2]][j], start + nr + which[[2]][j]],
+                              centre=sc[i + j, dim], level=conf.ellipses),
                       border="dark grey", lty="dashed", lwd=2)
       }
   }
 
   if(replicates) {
-      nt <- if(what == "both") nr + nc
-            else if(what == "rows") nr
-            else nc
-
       if(length(x$boot.results) > 0)
           pts <- x$boot.results$t
       else if(length(x$jack.results) > 0)
@@ -406,20 +514,20 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
                (layer - 1) * nd * (nr + nc) + c((dim[1] - 1) * (nr + nc), (dim[2] - 1) * (nr + nc)) + 1
 
       if(what %in% c("rows", "both")) {
-          points(pts[, seq.int(start[1], start[1] + nr - 1)],
-                 pts[, seq.int(start[2], start[2] + nr - 1)],
-                 pch=rep(rep(1:18, length.out=nt)[1:nr], each=npts),
-                 col=rep(rainbow(nt, alpha=0.5)[1:nr], each=npts),
+          points(pts[, seq.int(start[1], start[1] + nr - 1)[which[[1]]]],
+                 pts[, seq.int(start[2], start[2] + nr - 1)[which[[1]]]],
+                 pch=rep(rep(1:18, length.out=nsc)[1:nwr], each=npts),
+                 col=rep(rainbow(nsc, alpha=0.5)[1:nwr], each=npts),
                  cex=0.5)
 
           i <- nr
       }
 
       if(what %in% c("columns", "both")) {
-          points(pts[, nr + seq.int(start[1], start[1] + nc - 1)],
-                 pts[, nr + seq.int(start[2], start[2] + nc - 1)],
-                 pch=rep(rep(1:18, length.out=nt)[i:(i + nr)], each=npts),
-                 col=rep(rainbow(nt, alpha=0.5)[i + 1:nr], each=npts),
+          points(pts[, nr + seq.int(start[1], start[1] + nc - 1)[which[[2]]]],
+                 pts[, nr + seq.int(start[2], start[2] + nc - 1)[which[[2]]]],
+                 pch=rep(rep(1:18, length.out=nsc)[-(1:nwr)], each=npts),
+                 col=rep(rainbow(nsc, alpha=0.5)[-(1:nwr)], each=npts),
                  cex=0.5)
       }
   }
@@ -433,50 +541,128 @@ plot.assoc <- function(x, dim=c(1, 2), layer=1, what=c("both", "rows", "columns"
   }
 
   # If no diagonal-specific parameters are present, we use the association of the point to itself
-  dg <- x$diagonal
+  dg <- c(x$diagonal[which[[1]]], x$diagonal[which[[2]]])
   if(length(dg) == 0)
       dg <- rep(0, nsc)
 
   dg <- dg + sqrt(sc[,dim[1]]^2 + sc[,dim[2]]^2)
 
+
   # Draw small circles after bigger ones
   ord <- order(p, decreasing=TRUE)
   p <- p[ord]
-  sc <- sc[ord,]
+  sc <- sc[ord, , drop=FALSE]
   dg <- dg[ord]
   col <- col[ord]
+  pch <- pch[ord]
+  font <- font[ord]
   groups <- groups[ord]
 
   if(mass)
-      size <- p/mean(p)*cex
+      # sqrt() because symbols' areas are approximately equal to
+      # that of square whose side length is proportional to cex:
+      # we want the areas to be proportional, not the side lengths
+      size <- sqrt(p/mean(p))*cex
   else
       size <- cex
 
-  if(length(groups) == 0)
-      pch <- 21
-  else
-      pch <- rep(c(21, 24, 22, 23, 25), length.out=8)[groups]
-
   if(is.null(col))
       col <- rep("black", length.out=nsc)
 
   if(luminosity) {
       col <- rgb2hsv(col2rgb(col))
-      # 1.2 is here to ensure no point is drawn completely black or white
-      bg <- hsv(col["h",], col["s",], v=(1-pmin(1, abs(dg)/max(abs(dg)))/1.2))
-#       col <- grey(l=(1-pmin(1, abs(dg)/max)/1.2))
 
-      if(any(dg < 0))
+      # 1.2 is here to ensure no point is drawn completely black or white
+      v <- 1 - abs(dg)/max(c(0, abs(dg)), na.rm=TRUE)/1.2
+
+      # Supplementary points do not have diagonal values
+      v[!is.finite(v)] <- 0.5
+
+      bg <- hsv(col["h",], col["s",], v=v)
+
+      if(any(dg[!is.na(dg)] < 0))
           warning("Some diagonal parameters are negative. Corresponding points will be colored according to their absolute value.")
   }
   else {
       bg <- col
   }
 
-  points(sc[,dim], cex=size,
-         bg=bg, pch=pch, col="white")
+  # Draw white border for filled symbols that support it
+  border <- ifelse(pch %in% 21:25, "white", bg)
+
+  points(sc[, dim, drop=FALSE], cex=size,
+         bg=bg, pch=pch, col=border)
+
   box()
-  pointLabel(sc[,dim[1]], sc[,dim[2]], rownames(sc))
+
+  pointLabel(sc[, dim[1]], sc[, dim[2]], rownames(sc), font=font)
+}
+
+averaged.assoc <- function(x, type=c("average", "average.rotate")) {
+      type <- match.arg(type)
+
+      nd <- ncol(x$phi)
+      nr <- nrow(x$row)
+      nc <- nrow(x$col)
+
+      if(inherits(x, "assoc.symm")) {
+          p <- get.probs(x)$rp
+
+          phi <- colSums(sweep(x$phi, 1, prop.table(colSums(x$row.weights)), "*"))
+
+          if(type == "average")
+              return(list(phi=phi, row=x$row[,,1], col=x$col[,,1]))
+
+          adjsc <- sweep(x$row[,,1], 2, sqrt(phi), "*")
+
+          # Technique proposed in Goodman (1991), Appendix 4, but with eigenvalues decomposition
+          lambda <- matrix(0, nr, nc)
+          for(i in 1:nd)
+              lambda <- lambda + (adjsc[,i] %o% adjsc[,i])
+          lambda0 <- lambda * sqrt(p %o% p) # Eq. A.4.3
+          eigen <- eigen(lambda0, symmetric=TRUE)
+          sc2 <- diag(1/sqrt(p)) %*% eigen$vectors[,1:nd] # Eq. A.4.7
+          phi2 <- t(eigen$values[1:nd])
+
+          adjsc2 <- sweep(sc2, 2, sqrt(phi), "*")
+
+          rot <- procrustes(adjsc, adjsc2)$rot
+
+          row2 <- col2 <- adjsc2
+      }
+      else {
+          probs <- get.probs(x)
+          rp <- probs$rp
+          cp <- probs$cp
+
+          phi <- colSums(sweep(x$phi, 1, prop.table(colSums(x$row.weights)), "*"))
+
+          if(type == "average")
+              return(list(phi=phi, row=x$row[,,1], col=x$col[,,1]))
+
+          adjrow <- sweep(x$row[,,1], 2, sqrt(phi), "*")
+          adjcol <- sweep(x$col[,,1], 2, sqrt(phi), "*")
+
+          # Technique proposed in Goodman (1991), Appendix 4
+          lambda <- matrix(0, nr, nc)
+          for(i in 1:nd)
+              lambda <- lambda + adjrow[,i] %o% adjcol[,i]
+          lambda0 <- lambda * sqrt(rp %o% cp) # Eq. A.4.3
+          sv <- svd(lambda0)
+          row2 <- diag(1/sqrt(rp)) %*% sv$u[,1:nd] # Eq. A.4.7
+          col2 <- diag(1/sqrt(cp)) %*% sv$v[,1:nd] # Eq. A.4.7
+          phi2 <- t(sv$d[1:nd])
+
+          adjrow2 <- sweep(row2, 2, sqrt(phi2), "*")
+          adjcol2 <- sweep(col2, 2, sqrt(phi2), "*")
+          rot <- procrustes(rbind(sweep(adjrow, 1, rp, "*"), sweep(adjcol, 1, cp, "*")),
+                            rbind(sweep(adjrow2, 1, rp, "*"), sweep(adjcol2, 1, cp, "*")))$rot
+      }
+
+      rownames(row2) <- rownames(x$row)
+      rownames(col2) <- rownames(x$col)
+
+      list(rotation=rot, phi=phi2, row=row2, col=col2)
 }
 
 # Function taken from the directlabels package, but it is in the public domain
