@@ -1,10 +1,14 @@
 ## UNIDIFF model (Erikson & Goldthorpe, 1992), or uniform layer effect model (Xie, 1992)
 
 unidiff <- function(tab, diagonal=c("included", "excluded", "only"),
-                    constrain="auto", family=poisson,
+                    constrain="auto",
+                    weighting=c("marginal", "uniform", "none"), norm=2,
+                    family=poisson,
                     tolerance=1e-8, iterMax=5000,
-                    trace=FALSE, verbose=TRUE, ...) {
+                    trace=FALSE, verbose=TRUE,
+                    checkEstimability=TRUE, ...) {
   diagonal <- match.arg(diagonal)
+  weighting <- match.arg(weighting)
 
   tab <- as.table(tab)
 
@@ -19,11 +23,9 @@ unidiff <- function(tab, diagonal=c("included", "excluded", "only"),
   if(length(dim(tab)) > 3)
       tab <- margin.table(tab, 1:3)
 
-  # When gnm evaluates the formulas, tab will have been converted to a data.frame,
-  # with a fallback if both names are empty
-  vars <- make.names(names(dimnames(tab)))
-  if(length(vars) == 0)
-      vars <- c("Var1", "Var2", "Var3")
+  tab <- prepareTable(tab, TRUE)
+  vars <- names(dimnames(tab))
+
 
   f <- sprintf("Freq ~ %s + %s + %s + %s:%s + %s:%s",
                vars[1], vars[2], vars[3], vars[1], vars[3], vars[2], vars[3])
@@ -32,38 +34,30 @@ unidiff <- function(tab, diagonal=c("included", "excluded", "only"),
       f <- sprintf("%s + Mult(Exp(%s), %s:%s)", f, vars[3], vars[1], vars[2])
 
       if(identical(constrain, "auto"))
-          constrain <- sprintf("(Mult\\(Exp\\(.\\), \\Q%s:%s\\E\\)\\Q.%s%s\\E)|(Mult\\(Exp\\(\\Q%s\\E.*\\.(\\Q%s%s\\E:)|(:\\Q%s%s\\E$))",
-                               vars[1], vars[2], vars[3], dimnames(tab)[[3]][1],
-                               vars[3], vars[1], rownames(tab)[1], vars[2], colnames(tab)[1])
+          constrain <- sprintf("^Mult\\(Exp\\(.\\), \\Q%s:%s\\E\\)\\Q.%s%s\\E$",
+                               vars[1], vars[2], vars[3], dimnames(tab)[[3]][1])
   }
-  if(diagonal == "excluded") {
+  else if(diagonal == "excluded") {
       f <- sprintf("%s + %s:Diag(%s, %s) + Mult(Exp(%s), %s:%s)",
                    f, vars[3], vars[1], vars[2], vars[3], vars[1], vars[2])
 
       if(identical(constrain, "auto"))
-          # Last pattern matches diagonal coefficients
-          constrain <- sprintf("(Mult\\(Exp\\(.\\), \\Q%s:%s\\E\\)\\Q.%s%s\\E)|(Mult\\(Exp\\(\\Q%s\\E.*\\.(\\Q%s\\E)$)",
-                               vars[1], vars[2], vars[3], dimnames(tab)[[3]][1],
-                               vars[3], paste(paste(vars[1], rownames(tab), ":", vars[2],
-                                                    rownames(tab), sep=""), collapse="\\E|\\Q"))
+          constrain <- sprintf("^Mult\\(Exp\\(.\\), \\Q%s:%s\\E\\)\\Q.%s%s\\E$",
+                               vars[1], vars[2], vars[3], dimnames(tab)[[3]][1])
   }
   else if(diagonal == "only") {
       f <- sprintf("%s + Mult(Exp(%s), Diag(%s, %s))", f, vars[3], vars[1], vars[2])
 
       if(identical(constrain, "auto"))
-          constrain <- sprintf("(Mult\\(Exp\\(.\\), Diag\\(\\Q%s, %s\\E\\)\\)\\Q.%s%s\\E)|(Mult\\(Exp\\(\\Q%s\\E.*\\.(\\Q%s%s\\E:)|(:\\Q%s%s\\E$))",
-                               vars[1], vars[2], vars[3], dimnames(tab)[[3]][1],
-                               vars[3], vars[1], rownames(tab)[1], vars[2], colnames(tab)[1])
+          constrain <- sprintf("^Mult\\(Exp\\(.\\), Diag\\(\\Q%s, %s\\E\\)\\)\\Q.%s%s\\E$",
+                               vars[1], vars[2], vars[3], dimnames(tab)[[3]][1])
   }
 
-
-  # FIXME: we should be able to eliminate 3:1, but this triggers a "numerically singular system" error
-#  if(missing(eliminate))
-#      eliminate <- eval(parse(text=sprintf("quote(%s:%s)", vars[1], vars[3])))
+  eliminate <- eval(parse(text=sprintf("quote(%s:%s)", vars[1], vars[3])))
 
   # We need to handle ... manually, else they would not be found when modelFormula() evaluates the call
   args <- list(formula=as.formula(f), data=tab, constrain=constrain,
-               family=family, tolerance=tolerance, iterMax=iterMax,
+               family=family, eliminate=eliminate, tolerance=tolerance, iterMax=iterMax,
                trace=trace, verbose=verbose)
 
   model <- do.call("gnm", c(args, list(...)))
@@ -74,37 +68,69 @@ unidiff <- function(tab, diagonal=c("included", "excluded", "only"),
   model$unidiff <- list()
   
   model$unidiff$layer <- getContrasts(model, pickCoef(model, sprintf("Mult\\(Exp\\(\\.\\)", vars[3])),
-                                      "first", check=FALSE)
+                                      "first", check=checkEstimability)
 
 
+  nr <- nrow(tab)
+  nc <- ncol(tab)
+
   if(diagonal == "included") {
-      mat <- tab[,,1]
-      nr <- nrow(mat)
-      nc <- ncol(mat)
+      if(weighting == "marginal") {
+          p <- prop.table(tab)
+          rp <- margin.table(p, 1)
+          cp <- margin.table(p, 2)
+      }
+      else if(weighting == "none") {
+          rp <- rep(1, nr)
+          cp <- rep(1, nc)
+      }
+      else {
+          rp <- rep(1/nr, nr)
+          cp <- rep(1/nc, nc)
+      }
+
+      rp1 <- prop.table(rp)
+      cp1 <- prop.table(cp)
+
+      mat <- matrix(0, nr, nc)
       con <- matrix(0, length(coef(model)), length(mat))
       ind <- pickCoef(model, sprintf("Mult\\(Exp\\(\\Q%s\\E\\)", vars[3]))
       for(i in 1:nr) {
           for(j in 1:nc) {
               mat[] <- 0
-              mat[i,] <- -1/nc
-              mat[,j] <- -1/nr
-              mat[i,j] <- 1 - 1/nc - 1/nr
-              mat <- mat + 1/(nr * nc)
+              mat[i,] <- -cp1
+              mat[,j] <- -rp1
+              mat[i,j] <- 1 - rp1[i] - cp1[j]
+              mat <- mat + rp1 %o% cp1
               con[ind, (j - 1) * nr + i] <- mat
           }
       }
 
       colnames(con) <- names(ind)
-      model$unidiff$interaction <- gnm:::se(model, con)
+      model$unidiff$interaction <- gnm::se(model, con, checkEstimability=checkEstimability)
+
+      if(weighting == "marginal") {
+          model$unidiff$phi <- maor(fitted(model)[,,1], TRUE, norm=norm,
+                                    row.weights=rp, col.weights=cp)
+          model$unidiff$maor <- maor(fitted(model)[,,1], norm=norm,
+                                    row.weights=rp, col.weights=cp)
+      }
+      else {
+          model$unidiff$phi <- maor(fitted(model)[,,1], TRUE, norm=norm, weighting=weighting)
+          model$unidiff$maor <- maor(fitted(model)[,,1], norm=norm, weighting=weighting)
+      }
   }
   else if(diagonal == "only"){
+      if(weighting != "uniform")
+          warning("uniform weigthing is always used when diagonal=\"only\"")
+
       # Quasi-variances cannot be computed for these coefficients, so hide the warning
       # Also skip the reference level
       suppressMessages(model$unidiff$interaction <- getContrasts(model, pickCoef(model, sprintf("Mult\\(Exp\\(\\Q%s\\E\\)", vars[3])),
-                                                                 ref="first", check=TRUE)$qvframe[-1,])
+                                                                 ref="first", check=checkEstimability)$qvframe[-1,])
 
-     # Diag() sorts levels alphabetically, which is not practical
-     model$unidiff$interaction[c(1 + order(rownames(tab))),] <- model$unidiff$interaction
+      # Diag() sorts levels alphabetically, which is not practical
+      model$unidiff$interaction[c(1 + order(rownames(tab))),] <- model$unidiff$interaction
   }
   else {
      # Interaction coefficients cannot be identified for now
@@ -119,6 +145,7 @@ unidiff <- function(tab, diagonal=c("included", "excluded", "only"),
                                                   rownames(model$unidiff$interaction))
 
   model$unidiff$diagonal <- diagonal
+  model$unidiff$weighting <- weighting
 
   class(model) <- c("unidiff", class(model))
 
@@ -130,6 +157,16 @@ unidiff <- function(tab, diagonal=c("included", "excluded", "only"),
 print.unidiff <- function(x, digits=max(3, getOption("digits") - 4), ...) {
   cat("Call:\n", deparse(x$call), "\n", sep="", fill=TRUE)
 
+  cat("\nLayer coefficients:\n")
+  print(setNames(exp(x$unidiff$layer$qvframe$estimate), row.names(x$unidiff$layer$qvframe)),
+        digits=digits, print.gap=2, ...)
+
+  if(length(x$unidiff$phi) > 0) {
+      cat("\nLayer intrinsic association coefficients:\n")
+      print(setNames(exp(x$unidiff$layer$qvframe$estimate) * x$unidiff$phi, row.names(x$unidiff$layer$qvframe)),
+            digits=digits, print.gap=2, ...)
+  }
+
   if(x$unidiff$diagonal == "included") {
       cat("\nFull two-way interaction coefficients:\n")
       interaction <- x$data[,,1]
@@ -147,6 +184,8 @@ print.unidiff <- function(x, digits=max(3, getOption("digits") - 4), ...) {
 
   print.default(format(interaction, digits=digits, ...), quote=FALSE, print.gap=2)
 
+  cat("\nNormalization weights:", x$unidiff$weighting)
+
   printModelStats(x)
 
   invisible(x)
@@ -154,25 +193,35 @@ print.unidiff <- function(x, digits=max(3, getOption("digits") - 4), ...) {
 
 summary.unidiff <- function(object, ...) {
   layer <- object$unidiff$layer$qvframe[,-4]
+  phi <- object$unidiff$phi
   interaction <- object$unidiff$interaction
 
   layer <- cbind(exp(layer[,1]), layer, 2 * pnorm(-abs(layer[,1]/layer[,2])))
   colnames(layer) <- c("Exp(Estimate)", "Estimate", "Std. Error", "Quasi SE", "Pr(>|z|)")
   rownames(layer) <- paste(names(dimnames(object$data))[3], rownames(layer), sep="")
 
+  if(length(phi) > 0) {
+      phi.layer <- layer
+      phi.layer[,1] <- layer[,1] * phi
+      phi.layer[,2] <- layer[,2] + log(phi)
+      phi.layer[,3:4] <- layer[,3:4] * phi
+  }
+
   if(object$unidiff$diagonal != "excluded") {
       interaction <- cbind(interaction, 2 * pnorm(-abs(interaction[,1]/interaction[,2])))
       colnames(interaction) <- c("Estimate", "Std. Error", "Pr(>|z|)")
   }
 
-  res <- list(call=object$call, diagonal=object$unidiff$diagonal,
+  res <- list(call=object$call,
               deviance.resid=residuals(object, type="deviance"),
               chisq=sum(na.omit(c(residuals(object, "pearson")^2))),
               dissim=sum(na.omit(c(abs(residuals(object, "response")))))/sum(na.omit(c(abs(fitted(object)))))/2,
-              layer=layer, interaction=interaction,
+              layer=layer, phi.layer=phi.layer, interaction=interaction,
+              diagonal=object$unidiff$diagonal,
+              weighting=object$unidiff$weighting,
               deviance=object$deviance, df.residual=object$df.residual,
-              bic=extractAIC(object, k=log(sum(na.omit(c(object$data)))))[2],
-              aic=extractAIC(object)[2])
+              bic=object$deviance - log(sum(na.omit(c(object$data)))) * object$df.residual,
+              aic=object$deviance - 2 * object$df.residual)
 
   class(res) <- "summary.unidiff"
 
@@ -185,6 +234,11 @@ print.summary.unidiff <- function(x, digits=max(3, getOption("digits") - 4), ...
   cat("\nLayer coefficients:\n")
   printCoefmat(x$layer, digits, signif.legend=FALSE, print.gap=2, ...)
 
+  if(length(x$phi) > 0) {
+      cat("\nLayer phi association coefficients:\n")
+      printCoefmat(x$phi.layer, digits, signif.legend=FALSE, print.gap=2, ...)
+  }
+
   if(x$diagonal != "only")
       cat("\nFull two-way interaction coefficients:\n")
   else
@@ -196,22 +250,34 @@ print.summary.unidiff <- function(x, digits=max(3, getOption("digits") - 4), ...
       cat("Not supported.\n")
 
 
-  cat("\nDeviance:            ", format(x$deviance, digits),
-      "\nPearson chi-squared: ", format(x$chisq, digits),
-      "\nDissimilarity index: ", format(x$dissim * 100, digits), "%",
-      "\nResidual df:         ", x$df.residual,
-      "\nBIC:                 ", x$aic,
-      "\nAIC:                 ", x$bic, "\n", sep="")
+  cat("\nNormalization weights: ", x$weighting,
+      "\nDeviance:              ", format(x$deviance, digits),
+      "\nPearson chi-squared:   ", format(x$chisq, digits),
+      "\nDissimilarity index:   ", format(x$dissim * 100, digits), "%",
+      "\nResidual df:           ", x$df.residual,
+      "\nBIC:                   ", x$aic,
+      "\nAIC:                   ", x$bic, "\n", sep="")
 }
 
-plot.unidiff <- function(x, exponentiate=TRUE, se.type=c("quasi.se", "se"), conf.int=.95,
-                         numeric.auto=TRUE, type="o",
-                         xlab=names(dimnames(x$data))[3], ylab="Layer coefficient", ...) {
+plot.unidiff <- function(x, what=c("layer.coef", "phi", "maor"),
+                         se.type=c("quasi.se", "se"), conf.int=.95,
+                         numeric.auto=TRUE, type="p",
+                         xlab=names(dimnames(x$data))[3], ylab=NULL, add=FALSE, ...) {
   if(!inherits(x, "unidiff"))
       stop("x must be a unidiff object")
 
+  what <- match.arg(what)
   se.type <- match.arg(se.type)
 
+  if(is.null(ylab)) {
+      if(what == "layer.coef")
+          ylab <- "Log odds ratio coefficient"
+      else if(what == "phi")
+          ylab <- "Intrinsic association coefficient"
+      else
+          ylab <- "Mean absolute odds ratio"
+  }
+
   qv <- x$unidiff$layer$qvframe
 
   w <- qnorm((1 - conf.int)/2, lower.tail=FALSE)
@@ -227,10 +293,19 @@ plot.unidiff <- function(x, exponentiate=TRUE, se.type=c("quasi.se", "se"), conf
       tails <- qv$estimate - w * qv$SE
    }
 
-  if(exponentiate) {
-      coefs <- exp(coefs)
-      tops <- exp(tops)
-      tails <- exp(tails)
+  coefs <- exp(coefs)
+  tops <- exp(tops)
+  tails <- exp(tails)
+
+  if(what == "phi") {
+      coefs <- coefs * x$unidiff$phi
+      tops <- tops * x$unidiff$phi
+      tails <- tails * x$unidiff$phi
+  }
+  else if(what == "maor") {
+      coefs <- x$unidiff$maor^coefs
+      tops <- x$unidiff$maor^tops
+      tails <- x$unidiff$maor^tails
   }
 
   range <- max(tops) - min(tails)
@@ -243,9 +318,15 @@ plot.unidiff <- function(x, exponentiate=TRUE, se.type=c("quasi.se", "se"), conf
   else
       at <- factor(rownames(qv), levels=rownames(qv))
 
-  plot.default(at, coefs, type=type, xaxt="n",
-       ylim=ylim, xlab=xlab, ylab=ylab, ...)
-  axis(1, at, labels=at)
+  if(!isTRUE(add)) {
+      plot.default(at, coefs, type=type, xaxt="n",
+                   ylim=ylim, xlab=xlab, ylab=ylab, ...)
+      axis(1, at, labels=at)
+  }
+  else {
+      points.default(at, coefs, type=type, ...)
+  }
+
   segments(as.numeric(at), tails, as.numeric(at), tops)
 
   invisible(x)
